# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from fedn.network.grpc import fedn_pb2 as network_dot_grpc_dot_fedn__pb2

GRPC_GENERATED_VERSION = '1.66.2'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in network/grpc/fedn_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class ModelServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Upload = channel.stream_unary(
                '/fedn.ModelService/Upload',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ModelRequest.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.ModelResponse.FromString,
                _registered_method=True)
        self.Download = channel.unary_stream(
                '/fedn.ModelService/Download',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ModelRequest.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.ModelResponse.FromString,
                _registered_method=True)


class ModelServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Upload(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Download(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ModelServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Upload': grpc.stream_unary_rpc_method_handler(
                    servicer.Upload,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ModelRequest.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.ModelResponse.SerializeToString,
            ),
            'Download': grpc.unary_stream_rpc_method_handler(
                    servicer.Download,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ModelRequest.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.ModelResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'fedn.ModelService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('fedn.ModelService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class ModelService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Upload(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(
            request_iterator,
            target,
            '/fedn.ModelService/Upload',
            network_dot_grpc_dot_fedn__pb2.ModelRequest.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.ModelResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Download(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/fedn.ModelService/Download',
            network_dot_grpc_dot_fedn__pb2.ModelRequest.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.ModelResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class ControlStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Start = channel.unary_unary(
                '/fedn.Control/Start',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ControlRequest.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.ControlResponse.FromString,
                _registered_method=True)
        self.Stop = channel.unary_unary(
                '/fedn.Control/Stop',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ControlRequest.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.ControlResponse.FromString,
                _registered_method=True)
        self.FlushAggregationQueue = channel.unary_unary(
                '/fedn.Control/FlushAggregationQueue',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ControlRequest.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.ControlResponse.FromString,
                _registered_method=True)
        self.SetAggregator = channel.unary_unary(
                '/fedn.Control/SetAggregator',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ControlRequest.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.ControlResponse.FromString,
                _registered_method=True)
        self.SetServerFunctions = channel.unary_unary(
                '/fedn.Control/SetServerFunctions',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ControlRequest.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.ControlResponse.FromString,
                _registered_method=True)


class ControlServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Start(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Stop(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FlushAggregationQueue(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetAggregator(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetServerFunctions(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ControlServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Start': grpc.unary_unary_rpc_method_handler(
                    servicer.Start,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ControlRequest.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.ControlResponse.SerializeToString,
            ),
            'Stop': grpc.unary_unary_rpc_method_handler(
                    servicer.Stop,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ControlRequest.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.ControlResponse.SerializeToString,
            ),
            'FlushAggregationQueue': grpc.unary_unary_rpc_method_handler(
                    servicer.FlushAggregationQueue,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ControlRequest.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.ControlResponse.SerializeToString,
            ),
            'SetAggregator': grpc.unary_unary_rpc_method_handler(
                    servicer.SetAggregator,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ControlRequest.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.ControlResponse.SerializeToString,
            ),
            'SetServerFunctions': grpc.unary_unary_rpc_method_handler(
                    servicer.SetServerFunctions,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ControlRequest.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.ControlResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'fedn.Control', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('fedn.Control', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Control(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Start(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fedn.Control/Start',
            network_dot_grpc_dot_fedn__pb2.ControlRequest.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.ControlResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Stop(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fedn.Control/Stop',
            network_dot_grpc_dot_fedn__pb2.ControlRequest.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.ControlResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def FlushAggregationQueue(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fedn.Control/FlushAggregationQueue',
            network_dot_grpc_dot_fedn__pb2.ControlRequest.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.ControlResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SetAggregator(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fedn.Control/SetAggregator',
            network_dot_grpc_dot_fedn__pb2.ControlRequest.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.ControlResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SetServerFunctions(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fedn.Control/SetServerFunctions',
            network_dot_grpc_dot_fedn__pb2.ControlRequest.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.ControlResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class ReducerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetGlobalModel = channel.unary_unary(
                '/fedn.Reducer/GetGlobalModel',
                request_serializer=network_dot_grpc_dot_fedn__pb2.GetGlobalModelRequest.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.GetGlobalModelResponse.FromString,
                _registered_method=True)


class ReducerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetGlobalModel(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ReducerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetGlobalModel': grpc.unary_unary_rpc_method_handler(
                    servicer.GetGlobalModel,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.GetGlobalModelRequest.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.GetGlobalModelResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'fedn.Reducer', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('fedn.Reducer', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Reducer(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetGlobalModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fedn.Reducer/GetGlobalModel',
            network_dot_grpc_dot_fedn__pb2.GetGlobalModelRequest.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.GetGlobalModelResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class ConnectorStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.AllianceStatusStream = channel.unary_stream(
                '/fedn.Connector/AllianceStatusStream',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ClientAvailableMessage.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.Status.FromString,
                _registered_method=True)
        self.SendStatus = channel.unary_unary(
                '/fedn.Connector/SendStatus',
                request_serializer=network_dot_grpc_dot_fedn__pb2.Status.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.Response.FromString,
                _registered_method=True)
        self.ListActiveClients = channel.unary_unary(
                '/fedn.Connector/ListActiveClients',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ListClientsRequest.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.ClientList.FromString,
                _registered_method=True)
        self.AcceptingClients = channel.unary_unary(
                '/fedn.Connector/AcceptingClients',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ConnectionRequest.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.ConnectionResponse.FromString,
                _registered_method=True)
        self.SendHeartbeat = channel.unary_unary(
                '/fedn.Connector/SendHeartbeat',
                request_serializer=network_dot_grpc_dot_fedn__pb2.Heartbeat.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.Response.FromString,
                _registered_method=True)
        self.ReassignClient = channel.unary_unary(
                '/fedn.Connector/ReassignClient',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ReassignRequest.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.Response.FromString,
                _registered_method=True)
        self.ReconnectClient = channel.unary_unary(
                '/fedn.Connector/ReconnectClient',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ReconnectRequest.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.Response.FromString,
                _registered_method=True)


class ConnectorServicer(object):
    """Missing associated documentation comment in .proto file."""

    def AllianceStatusStream(self, request, context):
        """Stream endpoint for status updates
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendStatus(self, request, context):
        """Report endpoint
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListActiveClients(self, request, context):
        """rpc RegisterClient (ClientAvailableMessage) returns (Response);
        List active clients endpoint
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AcceptingClients(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendHeartbeat(self, request, context):
        """Client messaging to stay engaged.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ReassignClient(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ReconnectClient(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ConnectorServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'AllianceStatusStream': grpc.unary_stream_rpc_method_handler(
                    servicer.AllianceStatusStream,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ClientAvailableMessage.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.Status.SerializeToString,
            ),
            'SendStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.SendStatus,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.Status.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.Response.SerializeToString,
            ),
            'ListActiveClients': grpc.unary_unary_rpc_method_handler(
                    servicer.ListActiveClients,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ListClientsRequest.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.ClientList.SerializeToString,
            ),
            'AcceptingClients': grpc.unary_unary_rpc_method_handler(
                    servicer.AcceptingClients,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ConnectionRequest.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.ConnectionResponse.SerializeToString,
            ),
            'SendHeartbeat': grpc.unary_unary_rpc_method_handler(
                    servicer.SendHeartbeat,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.Heartbeat.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.Response.SerializeToString,
            ),
            'ReassignClient': grpc.unary_unary_rpc_method_handler(
                    servicer.ReassignClient,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ReassignRequest.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.Response.SerializeToString,
            ),
            'ReconnectClient': grpc.unary_unary_rpc_method_handler(
                    servicer.ReconnectClient,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ReconnectRequest.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.Response.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'fedn.Connector', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('fedn.Connector', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Connector(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def AllianceStatusStream(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/fedn.Connector/AllianceStatusStream',
            network_dot_grpc_dot_fedn__pb2.ClientAvailableMessage.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.Status.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fedn.Connector/SendStatus',
            network_dot_grpc_dot_fedn__pb2.Status.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.Response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListActiveClients(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fedn.Connector/ListActiveClients',
            network_dot_grpc_dot_fedn__pb2.ListClientsRequest.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.ClientList.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def AcceptingClients(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fedn.Connector/AcceptingClients',
            network_dot_grpc_dot_fedn__pb2.ConnectionRequest.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.ConnectionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendHeartbeat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fedn.Connector/SendHeartbeat',
            network_dot_grpc_dot_fedn__pb2.Heartbeat.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.Response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ReassignClient(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fedn.Connector/ReassignClient',
            network_dot_grpc_dot_fedn__pb2.ReassignRequest.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.Response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ReconnectClient(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fedn.Connector/ReconnectClient',
            network_dot_grpc_dot_fedn__pb2.ReconnectRequest.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.Response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class CombinerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.TaskStream = channel.unary_stream(
                '/fedn.Combiner/TaskStream',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ClientAvailableMessage.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.TaskRequest.FromString,
                _registered_method=True)
        self.SendModelUpdate = channel.unary_unary(
                '/fedn.Combiner/SendModelUpdate',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ModelUpdate.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.Response.FromString,
                _registered_method=True)
        self.SendModelValidation = channel.unary_unary(
                '/fedn.Combiner/SendModelValidation',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ModelValidation.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.Response.FromString,
                _registered_method=True)
        self.SendModelPrediction = channel.unary_unary(
                '/fedn.Combiner/SendModelPrediction',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ModelPrediction.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.Response.FromString,
                _registered_method=True)


class CombinerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def TaskStream(self, request, context):
        """Stream endpoints for training/validation pub/sub
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendModelUpdate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendModelValidation(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendModelPrediction(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_CombinerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'TaskStream': grpc.unary_stream_rpc_method_handler(
                    servicer.TaskStream,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ClientAvailableMessage.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.TaskRequest.SerializeToString,
            ),
            'SendModelUpdate': grpc.unary_unary_rpc_method_handler(
                    servicer.SendModelUpdate,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ModelUpdate.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.Response.SerializeToString,
            ),
            'SendModelValidation': grpc.unary_unary_rpc_method_handler(
                    servicer.SendModelValidation,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ModelValidation.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.Response.SerializeToString,
            ),
            'SendModelPrediction': grpc.unary_unary_rpc_method_handler(
                    servicer.SendModelPrediction,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ModelPrediction.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.Response.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'fedn.Combiner', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('fedn.Combiner', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Combiner(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def TaskStream(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/fedn.Combiner/TaskStream',
            network_dot_grpc_dot_fedn__pb2.ClientAvailableMessage.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.TaskRequest.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendModelUpdate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fedn.Combiner/SendModelUpdate',
            network_dot_grpc_dot_fedn__pb2.ModelUpdate.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.Response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendModelValidation(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fedn.Combiner/SendModelValidation',
            network_dot_grpc_dot_fedn__pb2.ModelValidation.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.Response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendModelPrediction(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fedn.Combiner/SendModelPrediction',
            network_dot_grpc_dot_fedn__pb2.ModelPrediction.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.Response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class FunctionServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.HandleProvidedFunctions = channel.unary_unary(
                '/fedn.FunctionService/HandleProvidedFunctions',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ProvidedFunctionsRequest.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.ProvidedFunctionsResponse.FromString,
                _registered_method=True)
        self.HandleClientConfig = channel.stream_unary(
                '/fedn.FunctionService/HandleClientConfig',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ClientConfigRequest.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.ClientConfigResponse.FromString,
                _registered_method=True)
        self.HandleClientSelection = channel.unary_unary(
                '/fedn.FunctionService/HandleClientSelection',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ClientSelectionRequest.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.ClientSelectionResponse.FromString,
                _registered_method=True)
        self.HandleMetadata = channel.unary_unary(
                '/fedn.FunctionService/HandleMetadata',
                request_serializer=network_dot_grpc_dot_fedn__pb2.ClientMetaRequest.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.ClientMetaResponse.FromString,
                _registered_method=True)
        self.HandleStoreModel = channel.stream_unary(
                '/fedn.FunctionService/HandleStoreModel',
                request_serializer=network_dot_grpc_dot_fedn__pb2.StoreModelRequest.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.StoreModelResponse.FromString,
                _registered_method=True)
        self.HandleAggregation = channel.unary_stream(
                '/fedn.FunctionService/HandleAggregation',
                request_serializer=network_dot_grpc_dot_fedn__pb2.AggregationRequest.SerializeToString,
                response_deserializer=network_dot_grpc_dot_fedn__pb2.AggregationResponse.FromString,
                _registered_method=True)


class FunctionServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def HandleProvidedFunctions(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def HandleClientConfig(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def HandleClientSelection(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def HandleMetadata(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def HandleStoreModel(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def HandleAggregation(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_FunctionServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'HandleProvidedFunctions': grpc.unary_unary_rpc_method_handler(
                    servicer.HandleProvidedFunctions,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ProvidedFunctionsRequest.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.ProvidedFunctionsResponse.SerializeToString,
            ),
            'HandleClientConfig': grpc.stream_unary_rpc_method_handler(
                    servicer.HandleClientConfig,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ClientConfigRequest.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.ClientConfigResponse.SerializeToString,
            ),
            'HandleClientSelection': grpc.unary_unary_rpc_method_handler(
                    servicer.HandleClientSelection,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ClientSelectionRequest.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.ClientSelectionResponse.SerializeToString,
            ),
            'HandleMetadata': grpc.unary_unary_rpc_method_handler(
                    servicer.HandleMetadata,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.ClientMetaRequest.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.ClientMetaResponse.SerializeToString,
            ),
            'HandleStoreModel': grpc.stream_unary_rpc_method_handler(
                    servicer.HandleStoreModel,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.StoreModelRequest.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.StoreModelResponse.SerializeToString,
            ),
            'HandleAggregation': grpc.unary_stream_rpc_method_handler(
                    servicer.HandleAggregation,
                    request_deserializer=network_dot_grpc_dot_fedn__pb2.AggregationRequest.FromString,
                    response_serializer=network_dot_grpc_dot_fedn__pb2.AggregationResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'fedn.FunctionService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('fedn.FunctionService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class FunctionService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def HandleProvidedFunctions(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fedn.FunctionService/HandleProvidedFunctions',
            network_dot_grpc_dot_fedn__pb2.ProvidedFunctionsRequest.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.ProvidedFunctionsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def HandleClientConfig(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(
            request_iterator,
            target,
            '/fedn.FunctionService/HandleClientConfig',
            network_dot_grpc_dot_fedn__pb2.ClientConfigRequest.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.ClientConfigResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def HandleClientSelection(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fedn.FunctionService/HandleClientSelection',
            network_dot_grpc_dot_fedn__pb2.ClientSelectionRequest.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.ClientSelectionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def HandleMetadata(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fedn.FunctionService/HandleMetadata',
            network_dot_grpc_dot_fedn__pb2.ClientMetaRequest.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.ClientMetaResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def HandleStoreModel(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(
            request_iterator,
            target,
            '/fedn.FunctionService/HandleStoreModel',
            network_dot_grpc_dot_fedn__pb2.StoreModelRequest.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.StoreModelResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def HandleAggregation(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/fedn.FunctionService/HandleAggregation',
            network_dot_grpc_dot_fedn__pb2.AggregationRequest.SerializeToString,
            network_dot_grpc_dot_fedn__pb2.AggregationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



"""This module provides classes for managing database connections and stores in a federated network environment.

Classes:
    StoreContainer: A container for various store instances.
    DatabaseConnection: A singleton class for managing database connections and stores.
"""

import pymongo
from pymongo.database import Database
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from fedn.common.config import get_network_config, get_statestore_config
from fedn.network.storage.statestore.stores.analytic_store import AnalyticStore, MongoDBAnalyticStore
from fedn.network.storage.statestore.stores.client_store import ClientStore, MongoDBClientStore, SQLClientStore
from fedn.network.storage.statestore.stores.combiner_store import CombinerStore, MongoDBCombinerStore, SQLCombinerStore
from fedn.network.storage.statestore.stores.model_store import ModelStore, MongoDBModelStore, SQLModelStore
from fedn.network.storage.statestore.stores.package_store import MongoDBPackageStore, PackageStore, SQLPackageStore
from fedn.network.storage.statestore.stores.prediction_store import MongoDBPredictionStore, PredictionStore, SQLPredictionStore
from fedn.network.storage.statestore.stores.round_store import MongoDBRoundStore, RoundStore, SQLRoundStore
from fedn.network.storage.statestore.stores.session_store import MongoDBSessionStore, SessionStore, SQLSessionStore
from fedn.network.storage.statestore.stores.status_store import MongoDBStatusStore, SQLStatusStore, StatusStore
from fedn.network.storage.statestore.stores.store import MyAbstractBase
from fedn.network.storage.statestore.stores.validation_store import MongoDBValidationStore, SQLValidationStore, ValidationStore


class StoreContainer:
    """A container for various store instances."""

    def __init__(  # noqa: PLR0913
        self,
        client_store: ClientStore,
        validation_store: ValidationStore,
        combiner_store: CombinerStore,
        status_store: StatusStore,
        prediction_store: PredictionStore,
        round_store: RoundStore,
        package_store: PackageStore,
        model_store: ModelStore,
        session_store: SessionStore,
        analytic_store: AnalyticStore,
    ) -> None:
        """Initialize the StoreContainer with various store instances."""
        self.client_store = client_store
        self.validation_store = validation_store
        self.combiner_store = combiner_store
        self.status_store = status_store
        self.prediction_store = prediction_store
        self.round_store = round_store
        self.package_store = package_store
        self.model_store = model_store
        self.session_store = session_store
        self.analytic_store = analytic_store


class DatabaseConnection:
    """Singleton class for managing database connections and stores."""

    _instance = None

    def __new__(cls, *, force_create_new: bool = False) -> "DatabaseConnection":
        """Create a new instance of DatabaseConnection or return the existing singleton instance.

        Args:
            force_create_new (bool): If True, a new instance will be created regardless of the singleton pattern. Only used for testing purpose.

        Returns:
            DatabaseConnection: A new instance if force_create_new is True, otherwise the existing singleton instance.

        """
        if cls._instance is None or force_create_new:
            obj = super(DatabaseConnection, cls).__new__(cls)
            obj._init_connection()
            cls._instance = obj

        return cls._instance

    def _init_connection(self) -> None:
        statestore_config = get_statestore_config()
        network_id = get_network_config()

        if statestore_config["type"] == "MongoDB":
            mdb: Database = self._setup_mongo(statestore_config, network_id)

            client_store = MongoDBClientStore(mdb, "network.clients")
            validation_store = MongoDBValidationStore(mdb, "control.validations")
            combiner_store = MongoDBCombinerStore(mdb, "network.combiners")
            status_store = MongoDBStatusStore(mdb, "control.status")
            prediction_store = MongoDBPredictionStore(mdb, "control.predictions")
            round_store = MongoDBRoundStore(mdb, "control.rounds")
            package_store = MongoDBPackageStore(mdb, "control.packages")
            model_store = MongoDBModelStore(mdb, "control.models")
            session_store = MongoDBSessionStore(mdb, "control.sessions")
            analytic_store = MongoDBAnalyticStore(mdb, "control.analytics")

        elif statestore_config["type"] in ["SQLite", "PostgreSQL"]:
            Session = self._setup_sql(statestore_config)  # noqa: N806

            client_store = SQLClientStore(Session)
            validation_store = SQLValidationStore(Session)
            combiner_store = SQLCombinerStore(Session)
            status_store = SQLStatusStore(Session)
            prediction_store = SQLPredictionStore(Session)
            round_store = SQLRoundStore(Session)
            package_store = SQLPackageStore(Session)
            model_store = SQLModelStore(Session)
            session_store = SQLSessionStore(Session)
            analytic_store = None
        else:
            raise ValueError("Unknown statestore type")

        self.sc = StoreContainer(
            client_store,
            validation_store,
            combiner_store,
            status_store,
            prediction_store,
            round_store,
            package_store,
            model_store,
            session_store,
            analytic_store,
        )

    def close(self) -> None:
        """Close the database connection."""
        pass

    def _setup_mongo(self, statestore_config: dict, network_id: str) -> "DatabaseConnection":
        mc = pymongo.MongoClient(**statestore_config["mongo_config"])
        mc.server_info()
        mdb: Database = mc[network_id]

        return mdb

    def _setup_sql(self, statestore_config: dict) -> "DatabaseConnection":
        if statestore_config["type"] == "SQLite":
            sqlite_config = statestore_config["sqlite_config"]
            dbname = sqlite_config["dbname"]
            engine = create_engine(f"sqlite:///{dbname}", echo=False)
        elif statestore_config["type"] == "PostgreSQL":
            postgres_config = statestore_config["postgres_config"]
            username = postgres_config["username"]
            password = postgres_config["password"]
            host = postgres_config["host"]
            port = postgres_config["port"]

            engine = create_engine(f"postgresql://{username}:{password}@{host}:{port}/fedn_db", echo=False)

        Session = sessionmaker(engine)  # noqa: N806

        MyAbstractBase.metadata.create_all(engine, checkfirst=True)

        return Session

    def get_stores(self) -> StoreContainer:
        """Get the StoreContainer instance."""
        return self.sc

    @property
    def client_store(self) -> ClientStore:
        return self.sc.client_store

    @property
    def validation_store(self) -> ValidationStore:
        return self.sc.validation_store

    @property
    def combiner_store(self) -> CombinerStore:
        return self.sc.combiner_store

    @property
    def status_store(self) -> StatusStore:
        return self.sc.status_store

    @property
    def prediction_store(self) -> PredictionStore:
        return self.sc.prediction_store

    @property
    def round_store(self) -> RoundStore:
        return self.sc.round_store

    @property
    def package_store(self) -> PackageStore:
        return self.sc.package_store

    @property
    def model_store(self) -> ModelStore:
        return self.sc.model_store

    @property
    def session_store(self) -> SessionStore:
        return self.sc.session_store

    @property
    def analytic_store(self) -> AnalyticStore:
        return self.sc.analytic_store



from typing import Any, Dict, List, Optional, Tuple

import pymongo
from pymongo.database import Database
from sqlalchemy import ForeignKey, String, func, select
from sqlalchemy.orm import Mapped, mapped_column

from fedn.network.storage.statestore.stores.store import MongoDBStore, MyAbstractBase, SQLStore, Store


class Prediction:
    def __init__(
        self, id: str, model_id: str, data: str, correlation_id: str, timestamp: str, prediction_id: str, meta: str, sender: dict = None, receiver: dict = None
    ):
        self.id = id
        self.model_id = model_id
        self.data = data
        self.correlation_id = correlation_id
        self.timestamp = timestamp
        self.prediction_id = prediction_id
        self.meta = meta
        self.sender = sender
        self.receiver = receiver


class PredictionStore(Store[Prediction]):
    pass


class MongoDBPredictionStore(MongoDBStore[Prediction]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection)

    def get(self, id: str) -> Prediction:
        """Get an entity by id
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the Prediction (property)
        return: The entity
        """
        response = super().get(id)
        return response

    def update(self, id: str, item: Prediction) -> bool:
        raise NotImplementedError("Update not implemented for PredictionStore")

    def add(self, item: Prediction) -> Tuple[bool, Any]:
        return super().add(item)

    def delete(self, id: str) -> bool:
        raise NotImplementedError("Delete not implemented for PredictionStore")

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, **kwargs) -> Dict[int, List[Prediction]]:
        """List entities
        param limit: The maximum number of entities to return
            type: int
            description: The maximum number of entities to return
        param skip: The number of entities to skip
            type: int
            description: The number of entities to skip
        param sort_key: The key to sort by
            type: str
            description: The key to sort by
        param sort_order: The order to sort by
            type: pymongo.DESCENDING
            description: The order to sort by
        return: A dictionary with the count and a list of entities
        """
        return super().list(limit, skip, sort_key or "timestamp", sort_order, **kwargs)


class PredictionModel(MyAbstractBase):
    __tablename__ = "predictions"

    correlation_id: Mapped[str]
    data: Mapped[Optional[str]]
    model_id: Mapped[Optional[str]] = mapped_column(ForeignKey("models.id"))
    receiver_name: Mapped[Optional[str]] = mapped_column(String(255))
    receiver_role: Mapped[Optional[str]] = mapped_column(String(255))
    sender_name: Mapped[Optional[str]] = mapped_column(String(255))
    sender_role: Mapped[Optional[str]] = mapped_column(String(255))
    timestamp: Mapped[str] = mapped_column(String(255))
    prediction_id: Mapped[str] = mapped_column(String(255))


def from_row(row: PredictionModel) -> Prediction:
    return {
        "id": row.id,
        "model_id": row.model_id,
        "data": row.data,
        "correlation_id": row.correlation_id,
        "timestamp": row.timestamp,
        "prediction_id": row.prediction_id,
        "sender": {"name": row.sender_name, "role": row.sender_role},
        "receiver": {"name": row.receiver_name, "role": row.receiver_role},
    }


class SQLPredictionStore(PredictionStore, SQLStore[Prediction]):
    def __init__(self, Session):
        super().__init__(Session)

    def get(self, id: str) -> Prediction:
        with self.Session() as session:
            stmt = select(Prediction).where(Prediction.id == id)
            item = session.scalars(stmt).first()

            if item is None:
                return None
            return from_row(item)

    def update(self, id: str, item: Prediction) -> bool:
        raise NotImplementedError("Update not implemented for PredictionStore")

    def add(self, item: Prediction) -> Tuple[bool, Any]:
        with self.Session() as session:
            sender = item["sender"] if "sender" in item else None
            receiver = item["receiver"] if "receiver" in item else None

            validation = PredictionModel(
                correlation_id=item.get("correlationId") or item.get("correlation_id"),
                data=item.get("data"),
                model_id=item.get("modelId") or item.get("model_id"),
                receiver_name=receiver.get("name") if receiver else None,
                receiver_role=receiver.get("role") if receiver else None,
                sender_name=sender.get("name") if sender else None,
                sender_role=sender.get("role") if sender else None,
                prediction_id=item.get("predictionId") or item.get("prediction_id"),
                timestamp=item.get("timestamp"),
            )

            session.add(validation)
            session.commit()

            return True, validation

    def delete(self, id: str) -> bool:
        raise NotImplementedError("Delete not implemented for PredictionStore")

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, **kwargs):
        with self.Session() as session:
            stmt = select(PredictionModel)

            for key, value in kwargs.items():
                if key == "_id":
                    key = "id"
                elif key == "sender.name":
                    key = "sender_name"
                elif key == "sender.role":
                    key = "sender_role"
                elif key == "receiver.name":
                    key = "receiver_name"
                elif key == "receiver.role":
                    key = "receiver_role"
                elif key == "correlationId":
                    key = "correlation_id"
                elif key == "modelId":
                    key = "model_id"

                stmt = stmt.where(getattr(PredictionModel, key) == value)

            if sort_key:
                _sort_order: str = "DESC" if sort_order == pymongo.DESCENDING else "ASC"
                _sort_key: str = sort_key

                if _sort_key == "_id":
                    _sort_key = "id"
                elif _sort_key == "sender.name":
                    _sort_key = "sender_name"
                elif _sort_key == "sender.role":
                    _sort_key = "sender_role"
                elif _sort_key == "receiver.name":
                    _sort_key = "receiver_name"
                elif _sort_key == "receiver.role":
                    _sort_key = "receiver_role"
                elif _sort_key == "correlationId":
                    _sort_key = "correlation_id"
                elif _sort_key == "modelId":
                    _sort_key = "model_id"

                if _sort_key in PredictionModel.__table__.columns:
                    sort_obj = (
                        PredictionModel.__table__.columns.get(_sort_key) if _sort_order == "ASC" else PredictionModel.__table__.columns.get(_sort_key).desc()
                    )

                    stmt = stmt.order_by(sort_obj)

            if limit:
                stmt = stmt.offset(skip or 0).limit(limit)
            elif skip:
                stmt = stmt.offset(skip)

            items = session.execute(stmt)

            result = []

            for item in items:
                (r,) = item

                result.append(from_row(r))

            return {"count": len(result), "result": result}

    def count(self, **kwargs):
        with self.Session() as session:
            stmt = select(func.count()).select_from(PredictionModel)

            for key, value in kwargs.items():
                if key == "sender.name":
                    key = "sender_name"
                elif key == "sender.role":
                    key = "sender_role"
                elif key == "receiver.name":
                    key = "receiver_name"
                elif key == "receiver.role":
                    key = "receiver_role"
                elif key == "correlationId":
                    key = "correlation_id"
                elif key == "modelId":
                    key = "model_id"

                stmt = stmt.where(getattr(PredictionModel, key) == value)

            count = session.scalar(stmt)

            return count



from typing import List, Optional

from sqlalchemy import ForeignKey, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from fedn.network.storage.statestore.stores.store import MyAbstractBase


class SessionConfigModel(MyAbstractBase):
    __tablename__ = "session_configs"

    aggregator: Mapped[str] = mapped_column(String(255))
    round_timeout: Mapped[int]
    buffer_size: Mapped[int]
    delete_models_storage: Mapped[bool]
    clients_required: Mapped[int]
    validate: Mapped[bool]
    helper_type: Mapped[str] = mapped_column(String(255))
    model_id: Mapped[str] = mapped_column(ForeignKey("models.id"))
    session: Mapped["SessionModel"] = relationship(back_populates="session_config")


class SessionModel(MyAbstractBase):
    __tablename__ = "sessions"

    name: Mapped[Optional[str]] = mapped_column(String(255))
    status: Mapped[str] = mapped_column(String(255))
    session_config_id: Mapped[str] = mapped_column(ForeignKey("session_configs.id"))
    session_config: Mapped["SessionConfigModel"] = relationship(back_populates="session")
    models: Mapped[List["ModelModel"]] = relationship(back_populates="session")


class ModelModel(MyAbstractBase):
    __tablename__ = "models"

    active: Mapped[bool] = mapped_column(default=False)
    parent_model: Mapped[Optional[str]] = mapped_column(String(255))
    name: Mapped[Optional[str]] = mapped_column(String(255))
    session_configs: Mapped[List["SessionConfigModel"]] = relationship()
    session_id: Mapped[Optional[str]] = mapped_column(ForeignKey("sessions.id"))
    session: Mapped[Optional["SessionModel"]] = relationship(back_populates="models")


class RoundConfigModel(MyAbstractBase):
    __tablename__ = "round_configs"

    aggregator: Mapped[str] = mapped_column(String(255))
    round_timeout: Mapped[int]
    buffer_size: Mapped[int]
    delete_models_storage: Mapped[bool]
    clients_required: Mapped[int]
    validate: Mapped[bool]
    helper_type: Mapped[str] = mapped_column(String(255))
    model_id: Mapped[Optional[str]] = mapped_column(ForeignKey("models.id"))
    session_id: Mapped[Optional[str]] = mapped_column(ForeignKey("sessions.id"))
    round: Mapped["RoundModel"] = relationship(back_populates="round_config")
    task: Mapped[str] = mapped_column(String(255))
    round_id: Mapped[str]
    rounds: Mapped[int]


class RoundDataModel(MyAbstractBase):
    __tablename__ = "round_data"

    time_commit: Mapped[Optional[float]]
    reduce_time_aggregate_model: Mapped[Optional[float]]
    reduce_time_fetch_model: Mapped[Optional[float]]
    reduce_time_load_model: Mapped[Optional[float]]
    round: Mapped["RoundModel"] = relationship(back_populates="round_data")


class RoundCombinerModel(MyAbstractBase):
    __tablename__ = "round_combiners"

    model_id: Mapped[str]
    name: Mapped[str] = mapped_column(String(255))
    round_id: Mapped[str]
    status: Mapped[str] = mapped_column(String(255))
    time_exec_training: Mapped[float]

    config__job_id: Mapped[str] = mapped_column(String(255))
    config_aggregator: Mapped[str] = mapped_column(String(255))
    config_buffer_size: Mapped[int]
    config_clients_required: Mapped[int]
    config_delete_models_storage: Mapped[bool]
    config_helper_type: Mapped[str] = mapped_column(String(255))
    config_model_id: Mapped[str] = mapped_column(ForeignKey("models.id"))
    config_round_id: Mapped[str]
    config_round_timeout: Mapped[int]
    config_rounds: Mapped[int]
    config_session_id: Mapped[str] = mapped_column(ForeignKey("sessions.id"))
    config_task: Mapped[str] = mapped_column(String(255))
    config_validate: Mapped[bool]

    data_aggregation_time_nr_aggregated_models: Mapped[Optional[int]]
    data_aggregation_time_time_model_aggregation: Mapped[Optional[float]]
    data_aggregation_time_time_model_load: Mapped[Optional[float]]
    data_nr_expected_updates: Mapped[Optional[int]]
    data_nr_required_updates: Mapped[Optional[int]]
    data_time_combination: Mapped[Optional[float]]
    data_timeout: Mapped[Optional[float]]

    parent_round_id: Mapped[str] = mapped_column(ForeignKey("rounds.id"))
    round: Mapped["RoundModel"] = relationship(back_populates="combiners")


class RoundModel(MyAbstractBase):
    __tablename__ = "rounds"

    round_id: Mapped[str] = mapped_column(unique=True)  # TODO: Add unique constraint. Does this work?
    status: Mapped[str] = mapped_column()

    round_config_id: Mapped[Optional[str]] = mapped_column(ForeignKey("round_configs.id"))
    round_config: Mapped[Optional["RoundConfigModel"]] = relationship(back_populates="round")
    round_data_id: Mapped[Optional[str]] = mapped_column(ForeignKey("round_data.id"))
    round_data: Mapped[Optional["RoundDataModel"]] = relationship(back_populates="round")
    combiners: Mapped[List["RoundCombinerModel"]] = relationship(back_populates="round", cascade="all, delete-orphan")



import pytest
import pymongo

import itertools
import datetime
import uuid

from fedn.network.storage.dbconnection import DatabaseConnection



@pytest.fixture
def test_combiners():
    start_date = datetime.datetime(2021, 1, 4, 1, 2, 4)
    # NOTE: The SQL version does not support the dicts as parameters  parent and config
    # TODO: Creat using Combiner class
    return [{"id":str(uuid.uuid4()), "committed_at":start_date - datetime.timedelta(days=1), "name":"test_combiner1",
             "parent":"localhost",
            # "config":{},
             "ip":"123:13:12:2", "fqdn":"", "port":8080,
                        "updated_at":start_date - datetime.timedelta(days=52), "address":"test_address"},
            {"id":str(uuid.uuid4()), "committed_at":start_date - datetime.timedelta(days=2), "name":"test_combiner2",
             "parent":"localhost",
            # "config":{},
             "ip":"123:13:12:2", "fqdn":"", "port":8080,
                        "updated_at":start_date - datetime.timedelta(days=12), "address":"test_address"},
            {"id":str(uuid.uuid4()), "committed_at":start_date - datetime.timedelta(days=8), "name":"test_combiner3",
             "parent":"localhost",
            # "config":{},
             "ip":"123:13:12:5", "fqdn":"", "port":8080,
                        "updated_at":start_date - datetime.timedelta(days=322), "address":"test_address"},
            {"id":str(uuid.uuid4()), "committed_at":start_date - datetime.timedelta(days=4), "name":"test_combiner4",
             "parent":"localhost",
            # "config":{},
             "ip":"123:13:12:4", "fqdn":"", "port":8080,
                        "updated_at":start_date - datetime.timedelta(days=23), "address":"test_address"},
            {"id":str(uuid.uuid4()), "committed_at":start_date - datetime.timedelta(days=51), "name":"test_combiner5",
             "parent":"localhost",
            # "config":{},
             "ip":"123:13:12:3", "fqdn":"", "port":8080,
                        "updated_at":start_date - datetime.timedelta(days=22), "address":"test_address"},
            {"id":str(uuid.uuid4()), "committed_at":start_date - datetime.timedelta(days=9), "name":"test_combiner6",
             "parent":"localhost",
            # "config":{},
             "ip":"123:13:12:3", "fqdn":"", "port":8080,
                        "updated_at":start_date - datetime.timedelta(days=24), "address":"test_address"},
            {"id":str(uuid.uuid4()), "committed_at":start_date - datetime.timedelta(days=3), "name":"test_combiner8",
             "parent":"localhost",
            # "config":{},
             "ip":"123:13:12:3", "fqdn":"", "port":8080,
                        "updated_at":start_date - datetime.timedelta(days=42), "address":"test_address"},
            {"id":str(uuid.uuid4()), "committed_at":start_date - datetime.timedelta(days=14), "name":"test_combiner7",
             "parent":"localhost",
            # "config":{},
             "ip":"123:13:12:2", "fqdn":"", "port":8080,
                        "updated_at":start_date - datetime.timedelta(days=12), "address":"test_address1"}]


@pytest.fixture
def db_connections_with_data(postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection, test_combiners):
    for c in test_combiners:
        res, _ = mongo_connection.combiner_store.add(c)
        assert res == True

    for c in test_combiners:
        res, _ = postgres_connection.combiner_store.add(c)
        assert res == True

    for c in test_combiners:
        res, _ = sql_connection.combiner_store.add(c)
        assert res == True

    yield [("postgres", postgres_connection), ("sqlite", sql_connection), ("mongo", mongo_connection)]

    # TODO:Clean up



@pytest.fixture
def options():
    sorting_keys = (#None, 
                        "name",
                        #"committed_at",
                        #"updated_at",
                        #"invalid_key"
                        ) 
    limits = (None, 0, 1, 2, 99)
    skips = (None, 0, 1, 2, 99)
    desc = (None, pymongo.DESCENDING, pymongo.ASCENDING)
    opt_kwargs = ({}, {"ip":"123:13:12:3"}, {"address":"test_address1"}, {"fqdn":"", "name":"test_combiner77"})

    return list(itertools.product(limits, skips, sorting_keys, desc, opt_kwargs))

class TestCombinerStore:

    def test_add_update_delete(self, postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection):
        pass

    def test_list(self, db_connections_with_data: list[tuple[str, DatabaseConnection]], options: list[tuple]):   
        for (name1, db_1), (name2, db_2) in zip(db_connections_with_data[1:], db_connections_with_data[:-1]):
            print("Running tests between databases {} and {}".format(name1, name2))
            for *opt,kwargs in options:
                res = db_1.combiner_store.list(*opt, **kwargs)
                count, gathered_combiners = res["count"], res["result"]

                res = db_2.combiner_store.list(*opt, **kwargs)
                count2, gathered_combiners2 = res["count"], res["result"]
                #TODO: The count is not equal to the number of clients in the list, but the number of clients returned by the query before skip and limit
                #It is not clear what is the intended behavior
                # assert(count == len(gathered_clients))
                # assert count == count2
                assert len(gathered_combiners) == len(gathered_combiners2)

                for i in range(len(gathered_combiners)):
                    #NOTE: id are not equal between the two databases, I think it is due to id being overwritten in the _id field
                    #assert gathered_combiners2[i]["id"] == gathered_combiners[i]["id"]
                    #TODO: committed_at is not equal between the two databases, one reades from init the other uses the current time
                    #assert gathered_combiners2[i]["committed_at"] == gathered_combiners[i]["committed_at"]
                    assert gathered_combiners2[i]["name"] == gathered_combiners[i]["name"]
                    assert gathered_combiners2[i]["parent"] == gathered_combiners[i]["parent"]
                    assert gathered_combiners2[i]["ip"] == gathered_combiners[i]["ip"]
                    assert gathered_combiners2[i]["fqdn"] == gathered_combiners[i]["fqdn"]
                    assert gathered_combiners2[i]["port"] == gathered_combiners[i]["port"]
                    #TODO: updated_at is not equal between the two databases, one reades from init the other uses the current time
                    #assert gathered_combiners2[i]["updated_at"] == gathered_combiners[i]["updated_atssert gathered_combiners2[i]["address"] == gathered_combiners[i]["address"]
                        

                        
    



"""Module implementing Repository for MinIO."""

import io
from typing import IO, List

from minio import Minio
from minio.error import InvalidResponseError
from urllib3.poolmanager import PoolManager

from fedn.common.log_config import logger
from fedn.network.storage.s3.base import RepositoryBase


class MINIORepository(RepositoryBase):
    """Class implementing Repository for MinIO."""

    client = None

    def __init__(self, config: dict) -> None:
        """Initialize object.

        :param config: Dictionary containing configuration for credentials and bucket names.
        :type config: dict
        """
        super().__init__()
        self.name = "MINIORepository"

        if config["storage_secure_mode"]:
            manager = PoolManager(num_pools=100, cert_reqs="CERT_NONE", assert_hostname=False)
            self.client = Minio(
                "{0}:{1}".format(config["storage_hostname"], config["storage_port"]),
                access_key=config["storage_access_key"],
                secret_key=config["storage_secret_key"],
                secure=config["storage_secure_mode"],
                http_client=manager,
            )
        else:
            self.client = Minio(
                "{0}:{1}".format(config["storage_hostname"], config["storage_port"]),
                access_key=config["storage_access_key"],
                secret_key=config["storage_secret_key"],
                secure=config["storage_secure_mode"],
            )

    def set_artifact(self, instance_name: str, instance: IO, bucket: str, is_file: bool = False) -> bool:
        """Set object with name instance_name.

        :param instance_name: The name of the object
        :type instance_name: str
        :param instance: The object
        :type instance: Any
        :param bucket: The bucket name
        :type bucket: str
        :param is_file: Whether the instance is a file, defaults to False
        :type is_file: bool, optional
        :return: True if the artifact was set successfully
        :rtype: bool
        """
        try:
            if is_file:
                self.client.fput_object(bucket, instance_name, instance)
            else:
                self.client.put_object(bucket, instance_name, io.BytesIO(instance), len(instance))
        except Exception as e:
            logger.error(f"Failed to upload artifact: {instance_name} to bucket: {bucket}. Error: {e}")
            raise Exception(f"Could not load data into bytes: {e}") from e

        return True

    def get_artifact(self, instance_name: str, bucket: str) -> bytes:
        """Retrieve object with name instance_name.

        :param instance_name: The name of the object to retrieve
        :type instance_name: str
        :param bucket: The bucket name
        :type bucket: str
        :return: The retrieved object
        :rtype: bytes
        """
        try:
            data = self.client.get_object(bucket, instance_name)
            return data.read()
        except Exception as e:
            logger.error(f"Failed to fetch artifact: {instance_name} from bucket: {bucket}. Error: {e}")
            raise Exception(f"Could not fetch data from bucket: {e}") from e
        finally:
            data.close()
            data.release_conn()

    def get_artifact_stream(self, instance_name: str, bucket: str) -> io.BytesIO:
        """Return a stream handler for object with name instance_name.

        :param instance_name: The name of the object
        :type instance_name: str
        :param bucket: The bucket name
        :type bucket: str
        :return: Stream handler for object instance_name
        :rtype: io.BytesIO
        """
        try:
            return self.client.get_object(bucket, instance_name)
        except Exception as e:
            logger.error(f"Failed to fetch artifact stream: {instance_name} from bucket: {bucket}. Error: {e}")
            raise Exception(f"Could not fetch data from bucket: {e}") from e

    def list_artifacts(self, bucket: str) -> List[str]:
        """List all objects in bucket.

        :param bucket: Name of the bucket
        :type bucket: str
        :return: A list of object names
        :rtype: List[str]
        """
        objects = []
        try:
            objs = self.client.list_objects(bucket)
            for obj in objs:
                objects.append(obj.object_name)
        except Exception as err:
            logger.error(f"Failed to list artifacts in bucket: {bucket}. Error: {err}")
            raise Exception(f"Could not list models in bucket: {bucket}") from err
        return objects

    def delete_artifact(self, instance_name: str, bucket: str) -> None:
        """Delete object with name instance_name from buckets.

        :param instance_name: The object name
        :type instance_name: str
        :param bucket: Buckets to delete from
        :type bucket: str
        """
        try:
            self.client.remove_object(bucket, instance_name)
        except InvalidResponseError as err:
            logger.error(f"Could not delete artifact: {instance_name}. Error: {err}")

    def create_bucket(self, bucket_name: str) -> None:
        """Create a new bucket. If bucket exists, do nothing.

        :param bucket_name: The name of the bucket
        :type bucket_name: str
        """
        logger.info(f"Creating bucket: {bucket_name}")

        try:
            if not self.client.bucket_exists(bucket_name):
                self.client.make_bucket(bucket_name)
        except InvalidResponseError as err:
            logger.error(f"Failed to create bucket: {bucket_name}. Error: {err}")
            raise



"""Connector class for assigning clients to the FEDn network via the discovery service (REST-API).

The Connector class is used by the Client class in fedn/network/clients/client.py.
Once assigned, the client will retrieve combiner assignment from the discovery service.
The discovery service will also add the client to the statestore.
"""

import enum
from typing import Optional, Tuple

import requests

from fedn.common.config import (
    FEDN_AUTH_REFRESH_TOKEN,
    FEDN_AUTH_REFRESH_TOKEN_URI,
    FEDN_AUTH_SCHEME,
    FEDN_CUSTOM_URL_PREFIX,
)
from fedn.common.log_config import logger

# Constants for HTTP status codes
HTTP_STATUS_OK = 200
HTTP_STATUS_NO_CONTENT = 204
HTTP_STATUS_BAD_REQUEST = 400
HTTP_STATUS_UNAUTHORIZED = 401

# Default timeout for requests
REQUEST_TIMEOUT = 10  # seconds


class Status(enum.Enum):
    """Enum for representing the status of a client assignment."""

    Unassigned = 0
    Assigned = 1
    TryAgain = 2
    UnAuthorized = 3
    UnMatchedConfig = 4


class ConnectorClient:
    """Connector for assigning client to a combiner in the FEDn network.

    :param host: host of discovery service
    :type host: str
    :param port: port of discovery service
    :type port: int
    :param token: token for authentication
    :type token: str
    :param name: name of client
    :type name: str
    :param remote_package: True if remote package is used, False if local
    :type remote_package: bool
    :param force_ssl: True if https is used, False if http
    :type force_ssl: bool
    :param verify: True if certificate is verified, False if not
    :type verify: bool
    :param combiner: name of preferred combiner
    :type combiner: Optional[str]
    :param id: id of client
    :type id: Optional[str]
    """

    def __init__(
        self,
        host: str,
        port: int,
        token: str,
        name: str,
        remote_package: bool,
        force_ssl: bool = False,
        verify: bool = False,
        combiner: Optional[str] = None,
        id: Optional[str] = None,
    ) -> None:
        """Initialize the ConnectorClient."""
        self.host = host
        self.port = port
        self.token = token
        self.name = name
        self.verify = verify
        self.preferred_combiner = combiner
        self.id = id
        self.package = "remote" if remote_package else "local"

        # for https we assume an ingress handles permanent redirect (308)
        self.prefix = "https://" if force_ssl else "http://"
        self.connect_string = f"{self.prefix}{self.host}:{self.port}" if self.port else f"{self.prefix}{self.host}"

        logger.info(f"Setting connection string to {self.connect_string}.")

    def assign(self) -> Tuple[Status, Optional[dict]]:
        """Connect client to FEDn network discovery service, ask for combiner assignment.

        :return: Tuple with assignment status, combiner connection information if successful, else None.
        :rtype: tuple(:class:`fedn.network.clients.connect.Status`, Optional[dict])
        """
        try:
            payload = {
                "name": self.name,
                "client_id": self.id,
                "preferred_combiner": self.preferred_combiner,
                "package": self.package,
            }
            retval = requests.post(
                self.connect_string + FEDN_CUSTOM_URL_PREFIX + "/add_client",
                json=payload,
                verify=self.verify,
                allow_redirects=True,
                headers={"Authorization": f"{FEDN_AUTH_SCHEME} {self.token}"},
                timeout=REQUEST_TIMEOUT,
            )
        except Exception as e:
            logger.debug(f"***** {e}")
            return Status.Unassigned, {}

        if retval.status_code == HTTP_STATUS_BAD_REQUEST:
            reason = retval.json()["message"]
            return Status.UnMatchedConfig, reason

        if retval.status_code == HTTP_STATUS_UNAUTHORIZED:
            reason = retval.json().get("message", "Unauthorized connection to reducer, make sure the correct token is set")
            logger.warning(reason)
            if reason == "Token expired":
                status_code = self.refresh_token()
                if HTTP_STATUS_OK <= status_code < HTTP_STATUS_NO_CONTENT:
                    logger.info("Token refreshed.")
                    return Status.TryAgain, reason
                return Status.UnAuthorized, "Could not refresh token"
            return Status.UnAuthorized, reason

        if HTTP_STATUS_OK <= retval.status_code < HTTP_STATUS_NO_CONTENT:
            if retval.json().get("status") == "retry":
                reason = retval.json().get("message", "Controller was not ready. Try again later.")
                return Status.TryAgain, reason

            return Status.Assigned, retval.json()

        return Status.Unassigned, None

    def refresh_token(self) -> int:
        """Refresh client token.

        :return: Status code of the token refresh request.
        :rtype: int
        """
        if not FEDN_AUTH_REFRESH_TOKEN_URI or not FEDN_AUTH_REFRESH_TOKEN:
            logger.error("No refresh token URI/Token set, cannot refresh token.")
            return HTTP_STATUS_UNAUTHORIZED

        payload = requests.post(
            FEDN_AUTH_REFRESH_TOKEN_URI,
            verify=self.verify,
            allow_redirects=True,
            json={"refresh": FEDN_AUTH_REFRESH_TOKEN},
            timeout=REQUEST_TIMEOUT,
        )
        self.token = payload.json()["access"]
        return payload.status_code



import pytest
import pymongo

import itertools
import datetime
import uuid

from fedn.network.storage.dbconnection import DatabaseConnection

@pytest.fixture
def test_packages():
        start_date = datetime.datetime(2021, 1, 4, 1, 2, 4)
        return [{"id":str(uuid.uuid4()), "committed_at":start_date+datetime.timedelta(days=266), "description":"Test package1", "file_name":"test_package.tar.gz", "helper":"numpyhelper", "name":"test_package1", "storage_file_name":"test_package.tar.gz", "active":True},
                {"id":str(uuid.uuid4()), "committed_at":start_date+datetime.timedelta(days=616), "description":"Test package1", "file_name":"test_package.tar.gz", "helper":"numpyhelper", "name":"test_package2", "storage_file_name":"test_package.tar.gz", "active":True},
                {"id":str(uuid.uuid4()), "committed_at":start_date+datetime.timedelta(days=176), "description":"Test package1", "file_name":"test_package.tar.gz", "helper":"numpyhelper", "name":"test_package5", "storage_file_name":"test_package.tar.gz", "active":True},
                {"id":str(uuid.uuid4()), "committed_at":start_date+datetime.timedelta(days=366), "description":"Test package2", "file_name":"test_package.tar.gz", "helper":"numpyhelper", "name":"test_package4", "storage_file_name":"test_package.tar.gz", "active":True},
                {"id":str(uuid.uuid4()), "committed_at":start_date+datetime.timedelta(days=664), "description":"Test package2", "file_name":"test_package.tar.gz", "helper":"numpyhelper", "name":"test_package7", "storage_file_name":"test_package.tar.gz", "active":True},
                {"id":str(uuid.uuid4()), "committed_at":start_date+datetime.timedelta(days=66), "description":"Test package2", "file_name":"test_package.tar.gz", "helper":"numpyhelper", "name":"test_package3", "storage_file_name":"test_package.tar.gz", "active":False},
                {"id":str(uuid.uuid4()), "committed_at":start_date+datetime.timedelta(days=166), "description":"Test package2", "file_name":"test_package.tar.gz", "helper":"numpyhelper", "name":"test_package6", "storage_file_name":"test_package.tar.gz", "active":True}]

@pytest.fixture
def db_connections_with_data(postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection, test_packages):
    for c in test_packages:
        res, _ = mongo_connection.package_store.add(c)
        assert res == True

    for c in test_packages:
        res, _ = postgres_connection.package_store.add(c)
        assert res == True

    for c in test_packages:
        res, _ = sql_connection.package_store.add(c)
        assert res == True

    yield [("postgres", postgres_connection), ("sqlite", sql_connection), ("mongo", mongo_connection)]

    # TODO:Clean up



@pytest.fixture
def options():
    sorting_keys = (#None, 
                        "name",
                        "committed_at",
                        #"invalid_key"
                        ) 
    limits = (None, 0, 1, 2, 99)
    skips = (None, 0, 1, 2, 99)
    desc = (None, pymongo.DESCENDING, pymongo.ASCENDING)
    opt_kwargs = ({}, {"description":"Test package2"}, 
                      #TODO: The active field is handled differently in the two databases, the value is not the same
                      #{"active":False}
                      )

    return list(itertools.product(limits, skips, sorting_keys, desc, opt_kwargs))

class TestPackageStore:

    def test_add_update_delete(self, postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection):
        pass
    
    def test_list(self, db_connections_with_data: list[tuple[str, DatabaseConnection]], options: list[tuple]):   
        for (name1, db_1), (name2, db_2) in zip(db_connections_with_data[1:], db_connections_with_data[:-1]):
            print("Running tests between databases {} and {}".format(name1, name2))
            for *opt,kwargs in options:
                res = db_1.package_store.list(*opt, **kwargs)
                count, gathered_packages = res["count"], res["result"]

                res = db_2.package_store.list(*opt, **kwargs)
                count2, gathered_packages2 = res["count"], res["result"]
                #TODO: The count is not equal to the number of clients in the list, but the number of clients returned by the query before skip and limit
                #It is not clear what is the intended behavior
                # assert(count == len(gathered_clients))
                # assert count == count2
                assert len(gathered_packages) == len(gathered_packages2)

                for i in range(len(gathered_packages)):
                    #NOTE: id are not equal between the two databases, I think it is due to id being overwritten in the _id field
                    #assert gathered_models2[i]["id"] == gathered_models[i]["id"]
                    assert gathered_packages2[i]["committed_at"] == gathered_packages[i]["committed_at"]
                    assert gathered_packages2[i]["name"] == gathered_packages[i]["name"]
                    assert gathered_packages2[i]["description"] == gathered_packages[i]["description"]
                    assert gathered_packages2[i]["file_name"] == gathered_packages[i]["file_name"]
                    assert gathered_packages2[i]["helper"] == gathered_packages[i]["helper"]
                    assert gathered_packages2[i]["storage_file_name"] == gathered_packages[i]["storage_file_name"]
                    assert gathered_packages2[i]["active"] == gathered_packages[i]["active"]

                    



import pytest
import pymongo

import datetime
import uuid
import itertools

from fedn.network.storage.dbconnection import DatabaseConnection




@pytest.fixture
def test_predictions():
    return [{"id":str(uuid.uuid4()), "correlation_id":"test_correlation_id1", "data":"test_data1", "model_id":None, 
             "receiver_name":"test_receiver_name1", "receiver_role":"test_receiver_role1", "sender_name":"test_sender_name1", 
             "sender_role":"test_sender_role1", "timestamp":"2024-13", "prediction_id":"test_prediction_id1"},
             {"id":str(uuid.uuid4()), "correlation_id":"test_correlation_id4", "data":"test_data1", "model_id":None, 
             "receiver_name":"test_receiver_name2", "receiver_role":"test_receiver_role1", "sender_name":"test_sender_name1", 
             "sender_role":"test_sender_role1", "timestamp":"2024-11", "prediction_id":"test_prediction_id1"},
             {"id":str(uuid.uuid4()), "correlation_id":"test_correlation_id3", "data":"test_data1", "model_id":None, 
             "receiver_name":"test_receiver_name3", "receiver_role":"test_receiver_role1", "sender_name":"test_sender_name2", 
             "sender_role":"test_sender_role1", "timestamp":"2024-12", "prediction_id":"test_prediction_id2"},
             {"id":str(uuid.uuid4()), "correlation_id":"test_correlation_id2", "data":"test_data1", "model_id":None, 
             "receiver_name":"test_receiver_name4", "receiver_role":"test_receiver_role1", "sender_name":"test_sender_name2", 
             "sender_role":"test_sender_role1", "timestamp":"2024-16", "prediction_id":"test_prediction_id2"}]

@pytest.fixture
def db_connections_with_data(postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection, test_predictions):
    for c in test_predictions:
        res, _ = mongo_connection.prediction_store.add(c)
        assert res == True

    for c in test_predictions:
        res, _ = postgres_connection.prediction_store.add(c)
        assert res == True

    for c in test_predictions:
        res, _ = sql_connection.prediction_store.add(c)
        assert res == True

    yield [("postgres", postgres_connection), ("sqlite", sql_connection), ("mongo", mongo_connection)]

    # TODO:Clean up



@pytest.fixture
def options():
    sorting_keys = (#None, 
                        "correlation_id", 
                        "timestamp", 
                        #"invalid_key"
                        ) 
    limits = (None, 0, 1, 2, 99)
    skips = (None, 0, 1, 2, 99)
    desc = (None, pymongo.DESCENDING, pymongo.ASCENDING)
    opt_kwargs = ({}, {"prediction_id":"test_prediction_id2"}, {"correlation_id":""})

    return list(itertools.product(limits, skips, sorting_keys, desc, opt_kwargs))

class TestPredictionStore:

    def test_add_update_delete(self, postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection):
        pass

    def test_list(self, db_connections_with_data: list[tuple[str, DatabaseConnection]], options: list[tuple]):   
        for (name1, db_1), (name2, db_2) in zip(db_connections_with_data[1:], db_connections_with_data[:-1]):
            print("Running tests between databases {} and {}".format(name1, name2))
            for *opt,kwargs in options:
                res = db_1.prediction_store.list(*opt, **kwargs)
                count, gathered_models = res["count"], res["result"]

                res = db_2.prediction_store.list(*opt, **kwargs)
                count2, gathered_models2 = res["count"], res["result"]
                #TODO: The count is not equal to the number of clients in the list, but the number of clients returned by the query before skip and limit
                #It is not clear what is the intended behavior
                # assert(count == len(gathered_clients))
                # assert count == count2
                assert len(gathered_models) == len(gathered_models2)

                for i in range(len(gathered_models)):
                    #NOTE: id are not equal between the two databases, I think it is due to id being overwritten in the _id field
                    #assert gathered_models2[i]["id"] == gathered_models[i]["id"]
                    assert gathered_models2[i]["correlation_id"] == gathered_models[i]["correlation_id"]
                    assert gathered_models2[i]["data"] == gathered_models[i]["data"]
                    assert gathered_models2[i]["model_id"] == gathered_models[i]["model_id"]
                    #TODO: Reciever and sender are handled differently in the two databases, the value is not the same
                    #assert gathered_models2[i]["receiver_name"] == gathered_models[i]["receiver_name"]
                    #assert gathered_models2[i]["receiver_role"] == gathered_models[i]["receiver_role"]
                    #assert gathered_models2[i]["sender_name"] == gathered_models[i]["sender_name"]
                    #assert gathered_models2[i]["sender_role"] == gathered_models[i]["sender_role"]
                    assert gathered_models2[i]["timestamp"] == gathered_models[i]["timestamp"]
                    assert gathered_models2[i]["prediction_id"] == gathered_models[i]["prediction_id"]




import os
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Tuple

import fedn.utils.helpers.helpers
from fedn.common.log_config import logger
from fedn.network.api.network import Network
from fedn.network.combiner.interfaces import CombinerUnavailableError
from fedn.network.combiner.roundhandler import RoundConfig
from fedn.network.state import ReducerState
from fedn.network.storage.s3.repository import Repository
from fedn.network.storage.statestore.stores.client_store import ClientStore
from fedn.network.storage.statestore.stores.combiner_store import CombinerStore
from fedn.network.storage.statestore.stores.model_store import ModelStore
from fedn.network.storage.statestore.stores.package_store import PackageStore
from fedn.network.storage.statestore.stores.round_store import RoundStore
from fedn.network.storage.statestore.stores.session_store import MongoDBSessionStore

# Maximum number of tries to connect to statestore and retrieve storage configuration
MAX_TRIES_BACKEND = os.getenv("MAX_TRIES_BACKEND", 10)


class UnsupportedStorageBackend(Exception):
    pass


class MisconfiguredStorageBackend(Exception):
    pass


class MisconfiguredHelper(Exception):
    pass


class ControlBase(ABC):
    """Base class and interface for a global controller.
        Override this class to implement a global training strategy (control).

    :param statestore: The statestore object.
    :type statestore: :class:`fedn.network.statestore.statestorebase.StateStoreBase`
    """

    @abstractmethod
    def __init__(
        self,
        network_id: str,
        session_store: MongoDBSessionStore,
        model_store: ModelStore,
        round_store: RoundStore,
        package_store: PackageStore,
        combiner_store: CombinerStore,
        client_store: ClientStore,
        model_repository: Repository,
    ):
        """Constructor."""
        self._state = ReducerState.setup

        self.session_store = session_store
        self.model_store = model_store
        self.round_store = round_store
        self.package_store = package_store
        self.network = Network(self, network_id, combiner_store, client_store)

        self.model_repository = model_repository

        self._state = ReducerState.idle

    @abstractmethod
    def round(self, config, round_number):
        pass

    @abstractmethod
    def reduce(self, combiners):
        pass

    def get_helper(self):
        """Get a helper instance from global config.

        :return: Helper instance.
        :rtype: :class:`fedn.utils.plugins.helperbase.HelperBase`
        """
        helper_type: str = None

        try:
            active_package = self.package_store.get_active()
            helper_type = active_package["helper"]
        except Exception:
            logger.error("Failed to get active helper")

        helper = fedn.utils.helpers.helpers.get_helper(helper_type)
        if not helper:
            raise MisconfiguredHelper("Unsupported helper type {}, please configure compute_package.helper !".format(helper_type))
        return helper

    def get_state(self):
        """Get the current state of the controller.

        :return: The current state.
        :rtype: :class:`fedn.network.state.ReducerState`
        """
        return self._state

    def idle(self):
        """Check if the controller is idle.

        :return: True if idle, False otherwise.
        :rtype: bool
        """
        if self._state == ReducerState.idle:
            return True
        else:
            return False

    def get_latest_round_id(self) -> int:
        return self.round_store.get_latest_round_id()

    def get_compute_package_name(self):
        """:return:"""
        definition = self.package_store.get_active()
        if definition:
            try:
                package_name = definition["storage_file_name"]
                return package_name
            except (IndexError, KeyError):
                logger.error("No context filename set for compute context definition")
                return None
        else:
            return None

    def set_compute_package(self, filename, path):
        """Persist the configuration for the compute package."""
        self.model_repository.set_compute_package(filename, path)

    def get_compute_package(self, compute_package=""):
        """:param compute_package:
        :return:
        """
        if compute_package == "":
            compute_package = self.get_compute_package_name()
        if compute_package:
            return self.model_repository.get_compute_package(compute_package)
        else:
            return None

    def set_session_status(self, session_id: str, status: str) -> Tuple[bool, Any]:
        """Set the round round stats.

        :param round_id: The round unique identifier
        :type round_id: str
        :param status: The status
        :type status: str
        """
        session = self.session_store.get(session_id)
        session["status"] = status
        updated, msg = self.session_store.update(session["id"], session)
        if not updated:
            raise Exception(msg)

    def get_session_status(self, session_id: str):
        """Get the status of a session.

        :param session_id: The session unique identifier
        :type session_id: str
        :return: The status
        :rtype: str
        """
        session = self.session_store.get(session_id)
        return session["status"]

    def set_session_config(self, session_id: str, config: dict) -> Tuple[bool, Any]:
        """Set the model id for a session.

        :param session_id: The session unique identifier
        :type session_id: str
        :param config: The session config
        :type config: dict
        """
        session = self.session_store.get(session_id)
        session["session_config"] = config
        updated, msg = self.session_store.update(session["id"], session)
        if not updated:
            raise Exception(msg)

    def create_round(self, round_data):
        """Initialize a new round in backend db."""
        self.round_store.add(round_data)

    def set_round_data(self, round_id: str, round_data: dict):
        """Set round data.

        :param round_id: The round unique identifier
        :type round_id: str
        :param round_data: The status
        :type status: dict
        """
        round = self.round_store.get(round_id)
        round["round_data"] = round_data
        updated, _ = self.round_store.update(round["id"], round)
        if not updated:
            raise Exception("Failed to update round")

    def set_round_status(self, round_id: str, status: str):
        """Set the round round stats.

        :param round_id: The round unique identifier
        :type round_id: str
        :param status: The status
        :type status: str
        """
        round = self.round_store.get(round_id)
        round["status"] = status
        updated, _ = self.round_store.update(round["id"], round)
        if not updated:
            raise Exception("Failed to update round")

    def set_round_config(self, round_id: str, round_config: RoundConfig):
        """Upate round in backend db.

        :param round_id: The round unique identifier
        :type round_id: str
        :param round_config: The round configuration
        :type round_config: dict
        """
        round = self.round_store.get(round_id)
        round["round_config"] = round_config
        updated, _ = self.round_store.update(round["id"], round)
        if not updated:
            raise Exception("Failed to update round")

    def request_model_updates(self, combiners):
        """Ask Combiner server to produce a model update.

        :param combiners: A list of combiners
        :type combiners: tuple (combiner, combiner_round_config)
        """
        cl = []
        for combiner, combiner_round_config in combiners:
            response = combiner.submit(combiner_round_config)
            cl.append((combiner, response))
        return cl

    def commit(self, model_id: str, model: dict = None, session_id: str = None, name: str = None):
        """Commit a model to the global model trail. The model commited becomes the lastest consensus model.

        :param model_id: Unique identifier for the model to commit.
        :type model_id: str (uuid)
        :param model: The model object to commit
        :type model: BytesIO
        :param session_id: Unique identifier for the session
        :type session_id: str
        """
        helper = self.get_helper()
        if model is not None:
            outfile_name = helper.save(model)
            logger.info("Saving model file temporarily to {}".format(outfile_name))
            logger.info("CONTROL: Uploading model to Minio...")
            model_id = self.model_repository.set_model(outfile_name, is_file=True)

            logger.info("CONTROL: Deleting temporary model file...")
            os.unlink(outfile_name)

        logger.info("Committing model {} to global model trail in statestore...".format(model_id))

        active_model: str = None

        try:
            active_model = self.model_store.get_active()
        except Exception:
            logger.info("No active model, adding...")

        parent_model = None
        if active_model and session_id:
            parent_model = active_model

        committed_at = datetime.now()

        updated, _ = self.model_store.add(
            {
                "key": "models",
                "model": model_id,
                "parent_model": parent_model,
                "session_id": session_id,
                "committed_at": committed_at,
                "name": name,
            }
        )

        if not updated:
            raise Exception("Failed to commit model to global model trail")

        self.model_store.set_active(model_id)

    def get_combiner(self, name):
        for combiner in self.network.get_combiners():
            if combiner.name == name:
                return combiner
        return None

    def get_participating_combiners(self, combiner_round_config):
        """Assemble a list of combiners able to participate in a round as
        descibed by combiner_round_config.
        """
        combiners = []
        for combiner in self.network.get_combiners():
            try:
                # Current gRPC endpoint only returns active clients (both trainers and validators)
                nr_active_clients = len(combiner.list_active_clients())
            except CombinerUnavailableError:
                self._handle_unavailable_combiner(combiner)
                continue

            clients_required = int(combiner_round_config["clients_required"])
            is_participating = self.evaluate_round_participation_policy(clients_required, nr_active_clients)
            if is_participating:
                combiners.append((combiner, combiner_round_config))
        return combiners

    def evaluate_round_participation_policy(self, clients_required: int, nr_active_clients: int) -> bool:
        """Evaluate policy for combiner round-participation.
        A combiner participates if it is responsive and reports enough
        active clients to participate in the round.
        """
        if clients_required <= nr_active_clients:
            return True
        else:
            return False

    def evaluate_round_start_policy(self, combiners: list):
        """Check if the policy to start a round is met.

        :param combiners: A list of combiners
        :type combiners: list
        :return: True if the round policy is mer, otherwise False
        :rtype: bool
        """
        if len(combiners) > 0:
            return True
        else:
            return False

    def evaluate_round_validity_policy(self, round):
        """Check if the round is valid.

        At the end of the round, before committing a model to the global model trail,
        we check if the round validity policy has been met. This can involve
        e.g. asserting that a certain number of combiners have reported in an
        updated model, or that criteria on model performance have been met.

        :param round: The round object
        :rtype round: dict
        :return: True if the policy is met, otherwise False
        :rtype: bool
        """
        model_ids = []
        for combiner in round["combiners"]:
            try:
                model_ids.append(combiner["model_id"])
            except KeyError:
                pass

        if len(model_ids) == 0:
            return False

        return True

    def state(self):
        """Get the current state of the controller.

        :return: The state
        :rype: str
        """
        return self._state



import uuid
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Dict, Generic, List, Tuple, TypeVar

import pymongo
from bson import ObjectId
from pymongo.database import Database
from sqlalchemy import MetaData
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from fedn.network.storage.statestore.stores.shared import from_document

T = TypeVar("T")


class Store(ABC, Generic[T]):
    @abstractmethod
    def get(self, id: str) -> T:
        pass

    @abstractmethod
    def update(self, id: str, item: T) -> Tuple[bool, Any]:
        pass

    @abstractmethod
    def add(self, item: T) -> Tuple[bool, Any]:
        pass

    @abstractmethod
    def delete(self, id: str) -> bool:
        pass

    @abstractmethod
    def list(
        self,
        limit: int,
        skip: int,
        sort_key: str,
        sort_order=pymongo.DESCENDING,
        **kwargs,
    ) -> Dict[int, List[T]]:
        pass

    @abstractmethod
    def count(self, **kwargs) -> int:
        pass


class MongoDBStore(Store[T], Generic[T]):
    def __init__(self, database: Database, collection: str):
        self.database = database
        self.collection = collection

    def get(self, id: str) -> T:
        """Get an entity by id
        param id: The id of the entity
            type: str
        return: The entity
        """
        if not ObjectId.is_valid(id):
            return None
        id_obj = ObjectId(id)
        document = self.database[self.collection].find_one({"_id": id_obj})
        if document is None:
            return None

        return from_document(document)

    def update(self, id: str, item: T) -> Tuple[bool, Any]:
        try:
            result = self.database[self.collection].update_one({"_id": ObjectId(id)}, {"$set": item})
            if result.modified_count == 1:
                document = self.database[self.collection].find_one({"_id": ObjectId(id)})
                return True, from_document(document)
            else:
                return False, "Entity not found"
        except Exception as e:
            return False, str(e)

    def add(self, item: T) -> Tuple[bool, Any]:
        try:
            result = self.database[self.collection].insert_one(item)
            id = result.inserted_id
            document = self.database[self.collection].find_one({"_id": id})
            return True, from_document(document)
        except Exception as e:
            return False, str(e)

    def delete(self, id: str) -> bool:
        result = self.database[self.collection].delete_one({"_id": ObjectId(id)})
        return result.deleted_count == 1

    def list(
        self,
        limit: int,
        skip: int,
        sort_key: str,
        sort_order=pymongo.DESCENDING,
        **kwargs,
    ) -> Dict[int, List[T]]:
        """List entities
        param limit: The maximum number of entities to return
            type: int
        param skip: The number of entities to skip
            type: int
        param sort_key: The key to sort by
            type: str
        param sort_order: The order to sort by
            type: pymongo.DESCENDING | pymongo.ASCENDING
        param kwargs: Additional query parameters
            type: dict
            example: {"key": "models"}
        return: A dictionary with the count and the result
        """
        cursor = self.database[self.collection].find(kwargs).sort(sort_key, sort_order).skip(skip or 0).limit(limit or 0)

        count = self.database[self.collection].count_documents(kwargs)

        result = [from_document(document) for document in cursor]

        return {"count": count, "result": result}

    def count(self, **kwargs) -> int:
        """Count entities
        param kwargs: Additional query parameters
            type: dict
            example: {"key": "models"}
        return: The count (int)
        """
        return self.database[self.collection].count_documents(kwargs)


class SQLStore(Store[T]):
    def __init__(self, Session):
        self.Session = Session


constraint_naming_conventions = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}


class Base(DeclarativeBase):
    metadata = MetaData(naming_convention=constraint_naming_conventions)


class MyAbstractBase(Base):
    __abstract__ = True

    id: Mapped[str] = mapped_column(primary_key=True, default=lambda: str(uuid.uuid4()))
    committed_at: Mapped[datetime] = mapped_column(default=datetime.now())



import threading

from flask import Blueprint, jsonify, request

from fedn.common.log_config import logger
from fedn.network.api.auth import jwt_auth_required
from fedn.network.api.shared import control, model_store, prediction_store
from fedn.network.api.v1.shared import api_version, get_post_data_to_kwargs, get_typed_list_headers

bp = Blueprint("prediction", __name__, url_prefix=f"/api/{api_version}/predictions")


@bp.route("/start", methods=["POST"])
@jwt_auth_required(role="admin")
def start_session():
    """Start a new prediction session.
    param: prediction_id: The session id to start.
    type: prediction_id: str
    param: rounds: The number of rounds to run.
    type: rounds: int
    """
    try:
        data = request.json if request.headers["Content-Type"] == "application/json" else request.form.to_dict()
        prediction_id: str = data.get("prediction_id")

        if not prediction_id or prediction_id == "":
            return jsonify({"message": "prediction_id is required"}), 400

        session_config = {"prediction_id": prediction_id}

        if data.get("model_id") is None:
            count = model_store.count()
            if count == 0:
                return jsonify({"message": "No models available"}), 400
        else:
              model_id = data.get("model_id")
              model = model_store.get(model_id)
              if model is None:
                return jsonify({"message": f"Model {model_id} not found"}), 404
              session_config["model_id"] = model_id

        threading.Thread(target=control.prediction_session, kwargs={"config": session_config}).start()

        return jsonify({"message": "Prediction session started"}), 200
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
        return jsonify({"message": "Failed to start prediction session"}), 500


@bp.route("/", methods=["GET"])
@jwt_auth_required(role="admin")
def get_predictions():
    """Get predictions
    Retrieves a list of predictions based on the provided parameters.
    By specifying a parameter in the url, you can filter the predictions based on that parameter,
    and the response will contain only the predictions that match the filter.
    ---
    tags:
        - Predictions
    parameters:
      - name: sender.name
        in: query
        required: false
        type: string
        description: Name of the sender
      - name: sender.role
        in: query
        required: false
        type: string
        description: Role of the sender
      - name: receiver.name
        in: query
        required: false
        type: string
        description: Name of the receiver
      - name: receiver.role
        in: query
        required: false
        type: string
        description: Role of the receiver
      - name: prediction_id
        in: query
        required: false
        type: string
        description: Prediction id of the prediction
      - name: model_id
        in: query
        required: false
        type: string
      - name: correlation_id
        in: query
        required: false
        type: string
        description: Correlation id of the prediction
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of predictions to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of predictions to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the predictions by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the predictions in ('asc' or 'desc')
    definitions:
      Prediction:
        type: object
        properties:
          id:
            type: string
          correlation_id:
            type: string
          prediction_id:
            type: string
          model_id:
            type: string
          timestamp:
            type: object
            format: date-time
          data:
            type: string
          meta:
            type: string
          sender:
            type: object
            properties:
                name:
                    type: string
                role:
                    type: string
          receiver:
            type: object
            properties:
                name:
                    type: string
                role:
                    type: string
    responses:
      200:
        description: A list of predictions and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Prediction'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                message:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order = get_typed_list_headers(request.headers)
        kwargs = request.args.to_dict()

        response = prediction_store.list(limit, skip, sort_key, sort_order, **kwargs)

        return jsonify(response), 200
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/list", methods=["POST"])
@jwt_auth_required(role="admin")
def list_predictions():
    """List predictions
    Retrieves a list of predictions based on the provided parameters.
    Works much like the GET predictions method, but allows for a more complex query.
    By specifying a parameter in the body, you can filter the predictions based on that parameter,
    and the response will contain only the predictions that match the filter. If the parameter value contains a comma,
    the filter will be an "in" query, meaning that the predictions will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
        - Predictions
    parameters:
      - name: prediction
        in: body
        required: false
        schema:
            type: object
            properties:
                sender.name:
                    type: string
                    description: Name of the sender
                sender.role:
                    required: false
                    type: string
                    description: Role of the sender
                receiver.name:
                    type: string
                    description: Name of the receiver
                receiver.role:
                    required: false
                    type: string
                    description: Role of the receiver
                prediction_id:
                    required: false
                    type: string
                model_id:
                    required: false
                    type: string
                correlation_id:
                    required: false
                    type: string
                    description: Correlation id of the status
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of predictions to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of predictions to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the predictions by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the predictions in ('asc' or 'desc')
    responses:
      200:
        description: A list of predictions and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Prediction'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                message:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order = get_typed_list_headers(request.headers)
        kwargs = get_post_data_to_kwargs(request)

        response = prediction_store.list(limit, skip, sort_key, sort_order, **kwargs)

        return jsonify(response), 200
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
        return jsonify({"message": "An unexpected error occurred"}), 500



from abc import ABC
from typing import Dict, List, Tuple

import numpy as np


class ServerFunctionsBase(ABC):
    """Base class that defines the structure for the Server Functions. Override these functions
    to add to the server workflow.
    """

    def __init__(self) -> None:
        """Initialize the ServerFunctionsBase class. This method can be overridden
        by subclasses if initialization logic is required.
        """
        pass

    def aggregate(self, previous_global: List[np.ndarray], client_updates: Dict[str, Tuple[List[np.ndarray], Dict]]) -> List[np.ndarray]:
        """Aggregates a list of parameters from clients.

        Args:
        ----
            previous_global (list[np.ndarray]): A list of parameters representing the global
            model from the previous round.

            client_updates (list[list[np.ndarray]]): A dictionary where the key is client ID,
            pointing to a tuple with the first element being client parameter and second element
            being the clients metadata.

        Returns:
        -------
            list[np.ndarray]: A list of numpy arrays representing the aggregated
            parameters across all clients.

        """
        pass

    def client_settings(self, global_model: List[np.ndarray]) -> Dict:
        """Returns metadata related to the model, which gets distributed to the clients.
        The dictionary may only contain primitive types.

        Args:
        ----
            global_model (list[np.ndarray]): A list of parameters representing the global
            model for the upcomming round.

        Returns:
        -------
            dict: A dictionary containing metadata information, supporting only primitive python types.

        """
        pass

    def client_selection(self, client_ids: List[str]) -> List:
        """Returns a list of client_id's of which clients to be used for the next training request.

        Args:
        ----
            client_ids (list[str]): A list of client_ids for all connected clients.

        Returns:
        -------
            list[str]: A list of client ids for which clients should be chosen for the next training round.

        """
        pass


# base implementation
class ServerFunctions(ServerFunctionsBase):
    pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: network/grpc/fedn.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'network/grpc/fedn.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17network/grpc/fedn.proto\x12\x04\x66\x65\x64n\x1a\x1fgoogle/protobuf/timestamp.proto\":\n\x08Response\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.fedn.Client\x12\x10\n\x08response\x18\x02 \x01(\t\"\xf1\x01\n\x06Status\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.fedn.Client\x12\x0e\n\x06status\x18\x02 \x01(\t\x12!\n\tlog_level\x18\x03 \x01(\x0e\x32\x0e.fedn.LogLevel\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\t\x12\x16\n\x0e\x63orrelation_id\x18\x05 \x01(\t\x12-\n\ttimestamp\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x1e\n\x04type\x18\x07 \x01(\x0e\x32\x10.fedn.StatusType\x12\r\n\x05\x65xtra\x18\x08 \x01(\t\x12\x12\n\nsession_id\x18\t \x01(\t\"\xd8\x01\n\x0bTaskRequest\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.fedn.Client\x12\x1e\n\x08receiver\x18\x02 \x01(\x0b\x32\x0c.fedn.Client\x12\x10\n\x08model_id\x18\x03 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\t\x12\x16\n\x0e\x63orrelation_id\x18\x05 \x01(\t\x12\x11\n\ttimestamp\x18\x06 \x01(\t\x12\x0c\n\x04meta\x18\x07 \x01(\t\x12\x12\n\nsession_id\x18\x08 \x01(\t\x12\x1e\n\x04type\x18\t \x01(\x0e\x32\x10.fedn.StatusType\"\xbf\x01\n\x0bModelUpdate\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.fedn.Client\x12\x1e\n\x08receiver\x18\x02 \x01(\x0b\x32\x0c.fedn.Client\x12\x10\n\x08model_id\x18\x03 \x01(\t\x12\x17\n\x0fmodel_update_id\x18\x04 \x01(\t\x12\x16\n\x0e\x63orrelation_id\x18\x05 \x01(\t\x12\x11\n\ttimestamp\x18\x06 \x01(\t\x12\x0c\n\x04meta\x18\x07 \x01(\t\x12\x0e\n\x06\x63onfig\x18\x08 \x01(\t\"\xd8\x01\n\x0fModelValidation\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.fedn.Client\x12\x1e\n\x08receiver\x18\x02 \x01(\x0b\x32\x0c.fedn.Client\x12\x10\n\x08model_id\x18\x03 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\t\x12\x16\n\x0e\x63orrelation_id\x18\x05 \x01(\t\x12-\n\ttimestamp\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0c\n\x04meta\x18\x07 \x01(\t\x12\x12\n\nsession_id\x18\x08 \x01(\t\"\xdb\x01\n\x0fModelPrediction\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.fedn.Client\x12\x1e\n\x08receiver\x18\x02 \x01(\x0b\x32\x0c.fedn.Client\x12\x10\n\x08model_id\x18\x03 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\t\x12\x16\n\x0e\x63orrelation_id\x18\x05 \x01(\t\x12-\n\ttimestamp\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0c\n\x04meta\x18\x07 \x01(\t\x12\x15\n\rprediction_id\x18\x08 \x01(\t\"\x89\x01\n\x0cModelRequest\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.fedn.Client\x12\x1e\n\x08receiver\x18\x02 \x01(\x0b\x32\x0c.fedn.Client\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x12\n\n\x02id\x18\x04 \x01(\t\x12!\n\x06status\x18\x05 \x01(\x0e\x32\x11.fedn.ModelStatus\"]\n\rModelResponse\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x12\n\n\x02id\x18\x02 \x01(\t\x12!\n\x06status\x18\x03 \x01(\x0e\x32\x11.fedn.ModelStatus\x12\x0f\n\x07message\x18\x04 \x01(\t\"U\n\x15GetGlobalModelRequest\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.fedn.Client\x12\x1e\n\x08receiver\x18\x02 \x01(\x0b\x32\x0c.fedn.Client\"h\n\x16GetGlobalModelResponse\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.fedn.Client\x12\x1e\n\x08receiver\x18\x02 \x01(\x0b\x32\x0c.fedn.Client\x12\x10\n\x08model_id\x18\x03 \x01(\t\")\n\tHeartbeat\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.fedn.Client\"W\n\x16\x43lientAvailableMessage\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.fedn.Client\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\t\x12\x11\n\ttimestamp\x18\x03 \x01(\t\"P\n\x12ListClientsRequest\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.fedn.Client\x12\x1c\n\x07\x63hannel\x18\x02 \x01(\x0e\x32\x0b.fedn.Queue\"*\n\nClientList\x12\x1c\n\x06\x63lient\x18\x01 \x03(\x0b\x32\x0c.fedn.Client\"C\n\x06\x43lient\x12\x18\n\x04role\x18\x01 \x01(\x0e\x32\n.fedn.Role\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x11\n\tclient_id\x18\x03 \x01(\t\"m\n\x0fReassignRequest\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.fedn.Client\x12\x1e\n\x08receiver\x18\x02 \x01(\x0b\x32\x0c.fedn.Client\x12\x0e\n\x06server\x18\x03 \x01(\t\x12\x0c\n\x04port\x18\x04 \x01(\r\"c\n\x10ReconnectRequest\x12\x1c\n\x06sender\x18\x01 \x01(\x0b\x32\x0c.fedn.Client\x12\x1e\n\x08receiver\x18\x02 \x01(\x0b\x32\x0c.fedn.Client\x12\x11\n\treconnect\x18\x03 \x01(\r\"\'\n\tParameter\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"T\n\x0e\x43ontrolRequest\x12\x1e\n\x07\x63ommand\x18\x01 \x01(\x0e\x32\r.fedn.Command\x12\"\n\tparameter\x18\x02 \x03(\x0b\x32\x0f.fedn.Parameter\"F\n\x0f\x43ontrolResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\"\n\tparameter\x18\x02 \x03(\x0b\x32\x0f.fedn.Parameter\"\x13\n\x11\x43onnectionRequest\"<\n\x12\x43onnectionResponse\x12&\n\x06status\x18\x01 \x01(\x0e\x32\x16.fedn.ConnectionStatus\"1\n\x18ProvidedFunctionsRequest\x12\x15\n\rfunction_code\x18\x01 \x01(\t\"\xac\x01\n\x19ProvidedFunctionsResponse\x12T\n\x13\x61vailable_functions\x18\x01 \x03(\x0b\x32\x37.fedn.ProvidedFunctionsResponse.AvailableFunctionsEntry\x1a\x39\n\x17\x41vailableFunctionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x08:\x02\x38\x01\"#\n\x13\x43lientConfigRequest\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"/\n\x14\x43lientConfigResponse\x12\x17\n\x0f\x63lient_settings\x18\x01 \x01(\t\",\n\x16\x43lientSelectionRequest\x12\x12\n\nclient_ids\x18\x01 \x01(\t\"-\n\x17\x43lientSelectionResponse\x12\x12\n\nclient_ids\x18\x01 \x01(\t\"8\n\x11\x43lientMetaRequest\x12\x10\n\x08metadata\x18\x01 \x01(\t\x12\x11\n\tclient_id\x18\x02 \x01(\t\"$\n\x12\x43lientMetaResponse\x12\x0e\n\x06status\x18\x01 \x01(\t\"-\n\x11StoreModelRequest\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x12\n\n\x02id\x18\x02 \x01(\t\"$\n\x12StoreModelResponse\x12\x0e\n\x06status\x18\x01 \x01(\t\"\'\n\x12\x41ggregationRequest\x12\x11\n\taggregate\x18\x01 \x01(\t\"#\n\x13\x41ggregationResponse\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c*\x98\x01\n\nStatusType\x12\x07\n\x03LOG\x10\x00\x12\x18\n\x14MODEL_UPDATE_REQUEST\x10\x01\x12\x10\n\x0cMODEL_UPDATE\x10\x02\x12\x1c\n\x18MODEL_VALIDATION_REQUEST\x10\x03\x12\x14\n\x10MODEL_VALIDATION\x10\x04\x12\x14\n\x10MODEL_PREDICTION\x10\x05\x12\x0b\n\x07NETWORK\x10\x06*L\n\x08LogLevel\x12\x08\n\x04NONE\x10\x00\x12\x08\n\x04INFO\x10\x01\x12\t\n\x05\x44\x45\x42UG\x10\x02\x12\x0b\n\x07WARNING\x10\x03\x12\t\n\x05\x45RROR\x10\x04\x12\t\n\x05\x41UDIT\x10\x05*$\n\x05Queue\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\x0e\n\nTASK_QUEUE\x10\x01*S\n\x0bModelStatus\x12\x06\n\x02OK\x10\x00\x12\x0f\n\x0bIN_PROGRESS\x10\x01\x12\x12\n\x0eIN_PROGRESS_OK\x10\x02\x12\n\n\x06\x46\x41ILED\x10\x03\x12\x0b\n\x07UNKNOWN\x10\x04*8\n\x04Role\x12\t\n\x05OTHER\x10\x00\x12\n\n\x06\x43LIENT\x10\x01\x12\x0c\n\x08\x43OMBINER\x10\x02\x12\x0b\n\x07REDUCER\x10\x03*J\n\x07\x43ommand\x12\x08\n\x04IDLE\x10\x00\x12\t\n\x05START\x10\x01\x12\t\n\x05PAUSE\x10\x02\x12\x08\n\x04STOP\x10\x03\x12\t\n\x05RESET\x10\x04\x12\n\n\x06REPORT\x10\x05*I\n\x10\x43onnectionStatus\x12\x11\n\rNOT_ACCEPTING\x10\x00\x12\r\n\tACCEPTING\x10\x01\x12\x13\n\x0fTRY_AGAIN_LATER\x10\x02\x32z\n\x0cModelService\x12\x33\n\x06Upload\x12\x12.fedn.ModelRequest\x1a\x13.fedn.ModelResponse(\x01\x12\x35\n\x08\x44ownload\x12\x12.fedn.ModelRequest\x1a\x13.fedn.ModelResponse0\x01\x32\xbb\x02\n\x07\x43ontrol\x12\x34\n\x05Start\x12\x14.fedn.ControlRequest\x1a\x15.fedn.ControlResponse\x12\x33\n\x04Stop\x12\x14.fedn.ControlRequest\x1a\x15.fedn.ControlResponse\x12\x44\n\x15\x46lushAggregationQueue\x12\x14.fedn.ControlRequest\x1a\x15.fedn.ControlResponse\x12<\n\rSetAggregator\x12\x14.fedn.ControlRequest\x1a\x15.fedn.ControlResponse\x12\x41\n\x12SetServerFunctions\x12\x14.fedn.ControlRequest\x1a\x15.fedn.ControlResponse2V\n\x07Reducer\x12K\n\x0eGetGlobalModel\x12\x1b.fedn.GetGlobalModelRequest\x1a\x1c.fedn.GetGlobalModelResponse2\xab\x03\n\tConnector\x12\x44\n\x14\x41llianceStatusStream\x12\x1c.fedn.ClientAvailableMessage\x1a\x0c.fedn.Status0\x01\x12*\n\nSendStatus\x12\x0c.fedn.Status\x1a\x0e.fedn.Response\x12?\n\x11ListActiveClients\x12\x18.fedn.ListClientsRequest\x1a\x10.fedn.ClientList\x12\x45\n\x10\x41\x63\x63\x65ptingClients\x12\x17.fedn.ConnectionRequest\x1a\x18.fedn.ConnectionResponse\x12\x30\n\rSendHeartbeat\x12\x0f.fedn.Heartbeat\x1a\x0e.fedn.Response\x12\x37\n\x0eReassignClient\x12\x15.fedn.ReassignRequest\x1a\x0e.fedn.Response\x12\x39\n\x0fReconnectClient\x12\x16.fedn.ReconnectRequest\x1a\x0e.fedn.Response2\xfd\x01\n\x08\x43ombiner\x12?\n\nTaskStream\x12\x1c.fedn.ClientAvailableMessage\x1a\x11.fedn.TaskRequest0\x01\x12\x34\n\x0fSendModelUpdate\x12\x11.fedn.ModelUpdate\x1a\x0e.fedn.Response\x12<\n\x13SendModelValidation\x12\x15.fedn.ModelValidation\x1a\x0e.fedn.Response\x12<\n\x13SendModelPrediction\x12\x15.fedn.ModelPrediction\x1a\x0e.fedn.Response2\xec\x03\n\x0f\x46unctionService\x12Z\n\x17HandleProvidedFunctions\x12\x1e.fedn.ProvidedFunctionsRequest\x1a\x1f.fedn.ProvidedFunctionsResponse\x12M\n\x12HandleClientConfig\x12\x19.fedn.ClientConfigRequest\x1a\x1a.fedn.ClientConfigResponse(\x01\x12T\n\x15HandleClientSelection\x12\x1c.fedn.ClientSelectionRequest\x1a\x1d.fedn.ClientSelectionResponse\x12\x43\n\x0eHandleMetadata\x12\x17.fedn.ClientMetaRequest\x1a\x18.fedn.ClientMetaResponse\x12G\n\x10HandleStoreModel\x12\x17.fedn.StoreModelRequest\x1a\x18.fedn.StoreModelResponse(\x01\x12J\n\x11HandleAggregation\x12\x18.fedn.AggregationRequest\x1a\x19.fedn.AggregationResponse0\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'network.grpc.fedn_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_PROVIDEDFUNCTIONSRESPONSE_AVAILABLEFUNCTIONSENTRY']._loaded_options = None
  _globals['_PROVIDEDFUNCTIONSRESPONSE_AVAILABLEFUNCTIONSENTRY']._serialized_options = b'8\001'
  _globals['_STATUSTYPE']._serialized_start=3138
  _globals['_STATUSTYPE']._serialized_end=3290
  _globals['_LOGLEVEL']._serialized_start=3292
  _globals['_LOGLEVEL']._serialized_end=3368
  _globals['_QUEUE']._serialized_start=3370
  _globals['_QUEUE']._serialized_end=3406
  _globals['_MODELSTATUS']._serialized_start=3408
  _globals['_MODELSTATUS']._serialized_end=3491
  _globals['_ROLE']._serialized_start=3493
  _globals['_ROLE']._serialized_end=3549
  _globals['_COMMAND']._serialized_start=3551
  _globals['_COMMAND']._serialized_end=3625
  _globals['_CONNECTIONSTATUS']._serialized_start=3627
  _globals['_CONNECTIONSTATUS']._serialized_end=3700
  _globals['_RESPONSE']._serialized_start=66
  _globals['_RESPONSE']._serialized_end=124
  _globals['_STATUS']._serialized_start=127
  _globals['_STATUS']._serialized_end=368
  _globals['_TASKREQUEST']._serialized_start=371
  _globals['_TASKREQUEST']._serialized_end=587
  _globals['_MODELUPDATE']._serialized_start=590
  _globals['_MODELUPDATE']._serialized_end=781
  _globals['_MODELVALIDATION']._serialized_start=784
  _globals['_MODELVALIDATION']._serialized_end=1000
  _globals['_MODELPREDICTION']._serialized_start=1003
  _globals['_MODELPREDICTION']._serialized_end=1222
  _globals['_MODELREQUEST']._serialized_start=1225
  _globals['_MODELREQUEST']._serialized_end=1362
  _globals['_MODELRESPONSE']._serialized_start=1364
  _globals['_MODELRESPONSE']._serialized_end=1457
  _globals['_GETGLOBALMODELREQUEST']._serialized_start=1459
  _globals['_GETGLOBALMODELREQUEST']._serialized_end=1544
  _globals['_GETGLOBALMODELRESPONSE']._serialized_start=1546
  _globals['_GETGLOBALMODELRESPONSE']._serialized_end=1650
  _globals['_HEARTBEAT']._serialized_start=1652
  _globals['_HEARTBEAT']._serialized_end=1693
  _globals['_CLIENTAVAILABLEMESSAGE']._serialized_start=1695
  _globals['_CLIENTAVAILABLEMESSAGE']._serialized_end=1782
  _globals['_LISTCLIENTSREQUEST']._serialized_start=1784
  _globals['_LISTCLIENTSREQUEST']._serialized_end=1864
  _globals['_CLIENTLIST']._serialized_start=1866
  _globals['_CLIENTLIST']._serialized_end=1908
  _globals['_CLIENT']._serialized_start=1910
  _globals['_CLIENT']._serialized_end=1977
  _globals['_REASSIGNREQUEST']._serialized_start=1979
  _globals['_REASSIGNREQUEST']._serialized_end=2088
  _globals['_RECONNECTREQUEST']._serialized_start=2090
  _globals['_RECONNECTREQUEST']._serialized_end=2189
  _globals['_PARAMETER']._serialized_start=2191
  _globals['_PARAMETER']._serialized_end=2230
  _globals['_CONTROLREQUEST']._serialized_start=2232
  _globals['_CONTROLREQUEST']._serialized_end=2316
  _globals['_CONTROLRESPONSE']._serialized_start=2318
  _globals['_CONTROLRESPONSE']._serialized_end=2388
  _globals['_CONNECTIONREQUEST']._serialized_start=2390
  _globals['_CONNECTIONREQUEST']._serialized_end=2409
  _globals['_CONNECTIONRESPONSE']._serialized_start=2411
  _globals['_CONNECTIONRESPONSE']._serialized_end=2471
  _globals['_PROVIDEDFUNCTIONSREQUEST']._serialized_start=2473
  _globals['_PROVIDEDFUNCTIONSREQUEST']._serialized_end=2522
  _globals['_PROVIDEDFUNCTIONSRESPONSE']._serialized_start=2525
  _globals['_PROVIDEDFUNCTIONSRESPONSE']._serialized_end=2697
  _globals['_PROVIDEDFUNCTIONSRESPONSE_AVAILABLEFUNCTIONSENTRY']._serialized_start=2640
  _globals['_PROVIDEDFUNCTIONSRESPONSE_AVAILABLEFUNCTIONSENTRY']._serialized_end=2697
  _globals['_CLIENTCONFIGREQUEST']._serialized_start=2699
  _globals['_CLIENTCONFIGREQUEST']._serialized_end=2734
  _globals['_CLIENTCONFIGRESPONSE']._serialized_start=2736
  _globals['_CLIENTCONFIGRESPONSE']._serialized_end=2783
  _globals['_CLIENTSELECTIONREQUEST']._serialized_start=2785
  _globals['_CLIENTSELECTIONREQUEST']._serialized_end=2829
  _globals['_CLIENTSELECTIONRESPONSE']._serialized_start=2831
  _globals['_CLIENTSELECTIONRESPONSE']._serialized_end=2876
  _globals['_CLIENTMETAREQUEST']._serialized_start=2878
  _globals['_CLIENTMETAREQUEST']._serialized_end=2934
  _globals['_CLIENTMETARESPONSE']._serialized_start=2936
  _globals['_CLIENTMETARESPONSE']._serialized_end=2972
  _globals['_STOREMODELREQUEST']._serialized_start=2974
  _globals['_STOREMODELREQUEST']._serialized_end=3019
  _globals['_STOREMODELRESPONSE']._serialized_start=3021
  _globals['_STOREMODELRESPONSE']._serialized_end=3057
  _globals['_AGGREGATIONREQUEST']._serialized_start=3059
  _globals['_AGGREGATIONREQUEST']._serialized_end=3098
  _globals['_AGGREGATIONRESPONSE']._serialized_start=3100
  _globals['_AGGREGATIONRESPONSE']._serialized_end=3135
  _globals['_MODELSERVICE']._serialized_start=3702
  _globals['_MODELSERVICE']._serialized_end=3824
  _globals['_CONTROL']._serialized_start=3827
  _globals['_CONTROL']._serialized_end=4142
  _globals['_REDUCER']._serialized_start=4144
  _globals['_REDUCER']._serialized_end=4230
  _globals['_CONNECTOR']._serialized_start=4233
  _globals['_CONNECTOR']._serialized_end=4660
  _globals['_COMBINER']._serialized_start=4663
  _globals['_COMBINER']._serialized_end=4916
  _globals['_FUNCTIONSERVICE']._serialized_start=4919
  _globals['_FUNCTIONSERVICE']._serialized_end=5411
# @@protoc_insertion_point(module_scope)
